/*
 *  Display_DC.i - C64 graphics display.
 *
 *  Frodo (C) 1994-1997,2002 Christian Bauer
 *  Dreamcast KOS stuff by Tolga ABACI
 */

#include <kos.h>
#include <math.h>

#include "C64.h"
#include "CIA.h"
#include "1541job.h"

/*
  C64 keyboard matrix:

    Bit 7   6   5   4   3   2   1   0
  0    CUD  F5  F3  F1  F7 CLR RET DEL
  1    SHL  E   S   Z   4   A   W   3
  2     X   T   F   C   6   D   R   5
  3     V   U   H   B   8   G   Y   7
  4     N   O   K   M   0   J   I   9
  5     ,   @   :   .   -   L   P   +
  6     /   ^   =  SHR HOM  ;   *   
  7    R/S  Q   C= SPC  2  CTL  <-  1*/

#define MATRIX(a,b) (((a) << 3) | (b))

static key_t row5_key_list[] = 
{
  {{ " ", 0 }, MATRIX(7, 4), 306.0f }
};

static key_t row4_key_list[] = 
{
  {{ "C=", 0 }, MATRIX(7, 5), 30.0f },
  {{ "SH", 0 }, MATRIX(1, 7), 45.0f },
  {{ "Z", 0}, MATRIX(1, 4), 30.0f },
  {{ "X", 0}, MATRIX(2, 7), 30.0f },
  {{ "C", 0}, MATRIX(2, 4), 30.0f },
  {{ "V", 0}, MATRIX(3, 7), 30.0f },
  {{ "B", 0}, MATRIX(3, 4), 30.0f },
  {{ "N", 0}, MATRIX(4, 7), 30.0f },
  {{ "M", 0}, MATRIX(4, 4), 30.0f },
  {{ ",", "<"}, MATRIX(5, 7), 30.0f },
  {{ ".", ">"}, MATRIX(5, 4), 30.0f },
  {{ "/", "?"}, MATRIX(6, 7), 30.0f },
  {{ "SH", 0}, MATRIX(6, 4), 45.0f },
  {{ "D", "U"}, MATRIX(0, 7), 30.0f },
  {{ "L", "R"}, MATRIX(0, 2), 30.0f }
};

static key_t row3_key_list[] = 
{
  {{ "S", "R" }, MATRIX(7, 7), 30.0f },
  {{ "L", "S" }, 0, 30.0f },
  {{ "A", 0}, MATRIX(1, 2), 30.0f },
  {{ "S", 0}, MATRIX(1, 5), 30.0f },
  {{ "D", 0}, MATRIX(2, 2), 30.0f },
  {{ "F", 0}, MATRIX(2, 5), 30.0f },
  {{ "G", 0}, MATRIX(3, 2), 30.0f },
  {{ "H", 0}, MATRIX(3, 5), 30.0f },
  {{ "J", 0}, MATRIX(4, 2), 30.0f },
  {{ "K", 0}, MATRIX(4, 5), 30.0f },
  {{ "L", 0}, MATRIX(5, 2), 30.0f },
  {{ ":", "["}, MATRIX(5, 5), 30.0f },
  {{ ";", "]"}, MATRIX(6, 2), 30.0f },
  {{ "=", 0}, MATRIX(6, 5), 30.0f },
  {{ "RET", 0}, MATRIX(0, 1), 60.0f }
};

static key_t row1_key_list[] = 
{
  {{ "<=", 0 }, MATRIX(7, 1), 30.0f },
  {{ "1", "!"}, MATRIX(7, 0), 30.0f },
  {{ "2", "\""}, MATRIX(7, 3), 30.0f },
  {{ "3", "#"}, MATRIX(1, 0), 30.0f },
  {{ "4", "$"}, MATRIX(1, 3), 30.0f },
  {{ "5", "%"}, MATRIX(2, 0), 30.0f },
  {{ "6", "&"}, MATRIX(2, 3), 30.0f },
  {{ "7", "'"}, MATRIX(3, 0), 30.0f },
  {{ "8", "( "}, MATRIX(3, 3), 30.0f },
  {{ "9", ")"}, MATRIX(4, 0), 30.0f },
  {{ "0", 0}, MATRIX(4, 3), 30.0f },
  {{ "+", 0}, MATRIX(5, 0), 30.0f },
  {{ "-", 0}, MATRIX(5, 3), 30.0f },
  {{ "", "f"}, MATRIX(6, 0), 30.0f },
  {{ "HOM", "CLR"}, MATRIX(6, 3), 30.0f },
  {{ "DEL", "INS"}, MATRIX(0, 0), 30.0f }
};

static key_t row2_key_list[] = 
{
  {{ "CTL", 0 }, MATRIX(7, 2), 47.0f },
  {{ "Q", 0}, MATRIX(7, 6), 30.0f },
  {{ "W", 0}, MATRIX(1, 1), 30.0f },
  {{ "E", 0}, MATRIX(1, 6), 30.0f },
  {{ "R", 0}, MATRIX(2, 1), 30.0f },
  {{ "T", 0}, MATRIX(2, 6), 30.0f },
  {{ "Y", 0}, MATRIX(3, 1), 30.0f },
  {{ "U", 0}, MATRIX(3, 6), 30.0f },
  {{ "I", 0}, MATRIX(4, 1), 30.0f },
  {{ "O", 0}, MATRIX(4, 6), 30.0f },
  {{ "P", 0}, MATRIX(5, 1), 30.0f },
  {{ "@", 0}, MATRIX(5, 6), 30.0f },
  {{ "*", 0}, MATRIX(6, 1), 30.0f },
  {{ "| ", 0}, MATRIX(6, 6), 30.0f },
  {{ "RST", 0}, 0, 47.0f }
};

static key_t rowf1_key_list[] = 
{
  {{ "f 1", 0 }, MATRIX(0, 4), 45.0f }
};

static key_t rowf3_key_list[] = 
{
  {{ "f 3", 0 }, MATRIX(0, 5), 45.0f }
};

static key_t rowf5_key_list[] = 
{
  {{ "f 5", 0 }, MATRIX(0, 6), 45.0f }
};

static key_t rowf7_key_list[] = 
{
  {{ "f 7", 0 }, MATRIX(0, 3), 45.0f }
};

static key_row_t key_row_list[] = 
{
  {
    { row1_key_list, 16 },
    30.f, 10.0f, 40.f, 4.0f,
    { &(key_row_list[4]), &(key_row_list[1]), &(key_row_list[5]), &(key_row_list[5]) }
  },
  {
    { row2_key_list, 15 },
    30.f, 54.0f, 40.f, 4.0f,
    { &(key_row_list[0]), &(key_row_list[2]), &(key_row_list[6]), &(key_row_list[6]) }
  },
  {
    { row3_key_list, 15 },
    15.f, 98.0f, 40.f, 4.0f,
    { &(key_row_list[1]), &(key_row_list[3]), &(key_row_list[7]), &(key_row_list[7]) }
  },
  {
    { row4_key_list, 15 },
    15.f, 142.0f, 40.f, 4.0f,
    { &(key_row_list[2]), &(key_row_list[4]), &(key_row_list[8]), &(key_row_list[8]) }
  },
  {
    { row5_key_list, 1 },
    100.f, 186.0f, 40.f, 4.0f,
    { &(key_row_list[3]), &(key_row_list[0]), 0, 0 }
  },
  {
    { rowf1_key_list, 1 },
    585.f, 10.0f, 40.f, 4.0f,
    { &(key_row_list[8]), &(key_row_list[6]), &(key_row_list[0]), &(key_row_list[0]) }
  },
  {
    { rowf3_key_list, 1 },
    585.f, 54.0f, 40.f, 4.0f,
    { &(key_row_list[5]), &(key_row_list[7]), &(key_row_list[1]), &(key_row_list[1]) }
  },
  {
    { rowf5_key_list, 1 },
    585.f, 98.0f, 40.f, 4.0f,
    { &(key_row_list[6]), &(key_row_list[8]), &(key_row_list[2]), &(key_row_list[2]) }
  },
  {
    { rowf7_key_list, 1 },
    585.f, 142.0f, 40.f, 4.0f,
    { &(key_row_list[7]), &(key_row_list[5]), &(key_row_list[3]), &(key_row_list[3]) }
  }
};

int C64Display::DreamcastKey2C64(int keycode)
{
  switch (keycode) 
  {
    //case KBD_KEY_asciicircum: return MATRIX(7,1);     // Tolga: Wat iz DiZ?
          
  case KBD_KEY_COMMA: return MATRIX(5,7);
  case KBD_KEY_PERIOD: return MATRIX(5,4);
		
  case KBD_KEY_A: return MATRIX(1,2);
  case KBD_KEY_B: return MATRIX(3,4);
  case KBD_KEY_C: return MATRIX(2,4);
  case KBD_KEY_D: return MATRIX(2,2);
  case KBD_KEY_E: return MATRIX(1,6);
  case KBD_KEY_F: return MATRIX(2,5);
  case KBD_KEY_G: return MATRIX(3,2);
  case KBD_KEY_H: return MATRIX(3,5);
  case KBD_KEY_I: return MATRIX(4,1);
  case KBD_KEY_J: return MATRIX(4,2);
  case KBD_KEY_K: return MATRIX(4,5);
  case KBD_KEY_L: return MATRIX(5,2);
  case KBD_KEY_M: return MATRIX(4,4);
  case KBD_KEY_N: return MATRIX(4,7);
  case KBD_KEY_O: return MATRIX(4,6);
  case KBD_KEY_P: return MATRIX(5,1);
  case KBD_KEY_Q: return MATRIX(7,6);
  case KBD_KEY_R: return MATRIX(2,1);
  case KBD_KEY_S: return MATRIX(1,5);
  case KBD_KEY_T: return MATRIX(2,6);
  case KBD_KEY_U: return MATRIX(3,6);
  case KBD_KEY_V: return MATRIX(3,7);
  case KBD_KEY_W: return MATRIX(1,1);
  case KBD_KEY_X: return MATRIX(2,7);
  case KBD_KEY_Y: return MATRIX(3,1);
  case KBD_KEY_Z: return MATRIX(1,4);
    
  case KBD_KEY_BACKSPACE: return MATRIX(0,0);
  case KBD_KEY_DEL: return MATRIX(0,0);
    //case KBD_KEY_LEFTCONTROL: return MATRIX(7,2);    // Tolga: FIX
  case KBD_KEY_TAB: return MATRIX(7,1);
  case KBD_KEY_ENTER: return MATRIX(0,1);
  case KBD_KEY_SPACE: return MATRIX(7,4);
    //case KBD_KEY_LEFTSHIFT: return MATRIX(1,7);      // Tolga: FIX
    //case KBD_KEY_RIGHTSHIFT: return MATRIX(6,4);     // Tolga: FIX
  case KBD_KEY_ESCAPE: return MATRIX(7,7);
    //case KBD_KEY_RIGHTCONTROL:
    //case KBD_KEY_LEFTALT:
    //case KBD_KEY_RIGHTALT: return MATRIX(7,5);       // Tolga: FIX

  case KBD_KEY_INSERT: return MATRIX(0,0) | 0x80;
  case KBD_KEY_HOME: return MATRIX(6,3);
  case KBD_KEY_END: return MATRIX(6,0);
  case KBD_KEY_PGUP: return MATRIX(6,6);
  case KBD_KEY_PGDOWN: return MATRIX(6,5);
		
  case KBD_KEY_UP: return MATRIX(0,7)| 0x80;
  case KBD_KEY_DOWN: return MATRIX(0,7);
  case KBD_KEY_LEFT: return MATRIX(0,2) | 0x80;
  case KBD_KEY_RIGHT: return MATRIX(0,2);
		
  case KBD_KEY_F1: return MATRIX(0,4);
  case KBD_KEY_F2: return MATRIX(0,4) | 0x80;
  case KBD_KEY_F3: return MATRIX(0,5);
  case KBD_KEY_F4: return MATRIX(0,5) | 0x80;
  case KBD_KEY_F5: return MATRIX(0,6);
  case KBD_KEY_F6: return MATRIX(0,6) | 0x80;
  case KBD_KEY_F7: return MATRIX(0,3);
  case KBD_KEY_F8: return MATRIX(0,3) | 0x80;
		
  case KBD_KEY_0: return MATRIX(4,3);
  case KBD_KEY_1: return MATRIX(7,0);
  case KBD_KEY_2: return MATRIX(7,3);
  case KBD_KEY_3: return MATRIX(1,0);
  case KBD_KEY_4: return MATRIX(1,3);
  case KBD_KEY_5: return MATRIX(2,0);
  case KBD_KEY_6: return MATRIX(2,3);
  case KBD_KEY_7: return MATRIX(3,0);
  case KBD_KEY_8: return MATRIX(3,3);
  case KBD_KEY_9: return MATRIX(4,0);

  case KBD_KEY_LBRACKET: return MATRIX(5,6);
  case KBD_KEY_RBRACKET: return MATRIX(6,1);
  case KBD_KEY_SLASH: return MATRIX(6,7);
  case KBD_KEY_SEMICOLON: return MATRIX(5,5);
    //case KBD_KEY_grave: return MATRIX(6,2);       // Tolga: ?
    //case KBD_KEY_numbersign: return MATRIX(6,5);  // Tolga: FIX
    //case KBD_KEY_ltgt: return MATRIX(6,6);
    //case KBD_KEY_MINUS: return MATRIX(5,0);
    //case KBD_KEY_equal: return MATRIX(5,3);
  default: return -1;
  }
}

C64Display::C64Display(C64 *the_c64) : TheC64(the_c64)
{
  x_size = 640;
  y_size = 480;

  anim_type = ANIM_NONE;
  anim_time = 0;

  zoom = 1.0f;
  omega = 0.0f;

  if (ThePrefs.PALMode)
    vid_set_mode(DM_640x480_PAL_IL, PM_RGB565);
  else 
    vid_set_mode(DM_640x480_NTSC_IL, PM_RGB565);

  kbd_set_queue(false);

  vic_buf = (uint8 *) malloc(DISPLAY_X * DISPLAY_Y);		 
  txr_buf[0] = (uint16*) memalign(32, 512 * 272 * 2);
  txr_buf[1] = (uint16*) memalign(32, 512 * 272 * 2);
  txr_buf[2] = (uint16*) memalign(32, 512 * 272 * 2);

  SetupPVR();

  start_hold = 5;
  keybd_shift = false;

  controller_create(&cntr, 0, 20);
  keyboard_create(&keybd, 0.0f, 240.0f, 640.0f, 240.0f, 10.0f, fs_fnt, fs_fcxt, key_row_list, 9);
}

C64Display::~C64Display()
{
  free(vic_buf);

  for (int i = 0; i < 3; i ++)
    free(txr_buf[i]);

  pvr_shutdown();
}

void C64Display::NewPrefs(Prefs *prefs)
{
  if (ThePrefs.PALMode != prefs->PALMode)
  {
    pvr_shutdown();
    
    if (prefs->PALMode)
      vid_set_mode(DM_640x480_PAL_IL, PM_RGB565);
    else 
      vid_set_mode(DM_640x480_NTSC_IL, PM_RGB565);
    
    SetupPVR();

    TheApp->TheC64->frame_count = 0;
  }
}

void C64Display::Speedometer(int speed)
{
  this->speed = speed;
}

// Taken from KOS Plasma Example
void C64Display::SetupPVR() 
{
  pvr_poly_cxt_t cxt;

  pvr_init_defaults();
  pvr_set_bg_color(0,0,0);

  txr[0] = pvr_mem_malloc(512 * 272 * 2);
  memset(txr[0], 0, 512 * 272 * 2);
  txr[1] = pvr_mem_malloc(512 * 272 * 2);
  memset(txr[1], 0, 512 * 272 * 2);

  pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_RGB565 | PVR_TXRFMT_NONTWIDDLED, 512, 256, txr[0], PVR_FILTER_BILINEAR);
  pvr_poly_compile(&hdr[0], &cxt);
  pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_RGB565 | PVR_TXRFMT_NONTWIDDLED, 512, 256, txr[1], PVR_FILTER_BILINEAR);
  pvr_poly_compile(&hdr[1], &cxt);
  pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_RGB565 | PVR_TXRFMT_NONTWIDDLED, 512, 16, (uint8*) txr[0] + 512 * 256 * 2, PVR_FILTER_BILINEAR);
  pvr_poly_compile(&hdr2[0], &cxt);
  pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_RGB565 | PVR_TXRFMT_NONTWIDDLED, 512, 16, (uint8*) txr[1] + 512 * 256 * 2, PVR_FILTER_BILINEAR);
  pvr_poly_compile(&hdr2[1], &cxt);

  txr_cur = 0;

  ss_fnt = plx_font_load("/rd/marblearch.txf");
  ss_fcxt = plx_fcxt_create(ss_fnt, PVR_LIST_TR_POLY);
  plx_fcxt_setsize(ss_fcxt, 18.0f);  

  fs_fnt = plx_font_load("/rd/courier.txf");
  fs_fcxt = plx_fcxt_create(fs_fnt, PVR_LIST_TR_POLY);
  plx_fcxt_setsize(fs_fcxt, 16.0f);  
}

void C64Display::CalcCoords()
{
  if (anim_type == ANIM_NONE)
  {
    x_ul = 0.0f;
    y_ul = 0.0f;
    x_ll = 0.0f;
    y_ll = (float) y_size;
    x_ur = (float) x_size;
    y_ur = 0.0f;
    x_lr = (float) x_size;
    y_lr = (float) y_size;    
    x_ml = 0.0f;
    y_ml = (float) y_size * 256.0f / 272.0f;
    x_mr = (float) x_size;
    y_mr = (float) y_size * 256.0f / 272.0f;
  }
  else 
  {  
    float half_x = x_size / 2.0f;
    float half_y = y_size / 2.0f;

    float sin_omega = sin(omega);
    float cos_omega = cos(omega);

    float hco = half_x * cos_omega * zoom;
    float hso = half_x * sin_omega * zoom;

    x_ul = half_x - hco + hso;
    y_ul = half_y - hso - hco;

    x_ll = half_x - hco - hso;
    y_ll = half_y - hso + hco;

    x_lr = half_x + hco - hso;
    y_lr = half_y + hso + hco;

    x_ur = half_x + hco + hso;
    y_ur = half_y + hso - hco;

    x_ml = half_x - hco - hso * 120.0f / 136.0f;
    y_ml = half_y - hso + hco * 120.0f / 136.0f;

    x_mr = half_x + hco - hso * 120.0f / 136.0f;;
    y_mr = half_y + hso + hco * 120.0f / 136.0f;;
  }
}

void C64Display::Animate()
{
  if (anim_type == ANIM_TURN)
  {
    omega += (20.0f * M_PI / 180.0f);
    zoom -= 0.03;

    anim_time --;
    
    if (anim_time == 0)
    {
      // Hackish

      TheApp->TheC64->QuitThyself();
      anim_type = ANIM_NONE;
      omega = 0.0f;
      zoom = 1.0f;
    }
  }
}

void C64Display::SendToPVR(uint32 color) 
{
  pvr_vertex_t vert;

  if (ThePrefs.ShowBorders)
  {
    pvr_prim(&hdr[txr_cur], sizeof(*hdr));
  
    vert.flags = PVR_CMD_VERTEX;
    vert.x = x_ml;
    vert.y = y_ml;
    vert.z = 1.0f;
    vert.u = 0.0f;
    vert.v = 1.0f; 
    vert.argb = color;
    vert.oargb = 0;
    pvr_prim(&vert, sizeof(vert));

    vert.x = x_ul;
    vert.y = y_ul;
    vert.v = 0.0f;    
    pvr_prim(&vert, sizeof(vert));

    vert.x = x_mr;
    vert.y = y_mr;    
    vert.u = 0.75f;
    vert.v = 1.0f; 
    pvr_prim(&vert, sizeof(vert));
  
    vert.flags = PVR_CMD_VERTEX_EOL;
    vert.x = x_ur;
    vert.y = y_ur;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(vert));

    pvr_prim(&hdr2[txr_cur], sizeof(*hdr2));
  
    vert.flags = PVR_CMD_VERTEX;
    vert.x = x_ll;
    vert.y = y_ll;
    vert.z = 1.0f;
    vert.u = 0.0f;
    vert.v = 1.0f; 
    vert.argb = color;
    vert.oargb = 0;
    pvr_prim(&vert, sizeof(vert));
    
    vert.x = x_ml;    
    vert.y = y_ml;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(vert));
  
    vert.x = x_lr;
    vert.y = y_lr;
    vert.u = 0.75f;
    vert.v = 1.0f; 
    pvr_prim(&vert, sizeof(vert));
  
    vert.flags = PVR_CMD_VERTEX_EOL;
    vert.x = x_mr;
    vert.y = y_mr;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(vert));

  }
  else
  {
    pvr_prim(&hdr[txr_cur], sizeof(*hdr));
  
    vert.flags = PVR_CMD_VERTEX;
    vert.x = x_ll;
    vert.y = y_ll;
    vert.z = 1.0f;
    vert.u = 0.0f;
    vert.v = 0.9375f; 
    vert.argb = color;
    vert.oargb = 0;
    pvr_prim(&vert, sizeof(vert));
    
    vert.x = x_ul;
    vert.y = y_ul;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(vert));
  
    vert.x = x_lr;
    vert.y = y_lr;
    vert.u = 0.625f;
    vert.v = 0.9375f; 
    pvr_prim(&vert, sizeof(vert));
  
    vert.flags = PVR_CMD_VERTEX_EOL;
    vert.x = x_ur;
    vert.y = y_ur;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(vert));

  }
}

void C64Display::RenderToTexture()
{
  uint32 *p_fb;
  uint16 *p_vb, *vb_end;  

  if (ThePrefs.ShowBorders)
  {
    p_fb = (uint32*) txr_buf[txr_cur];
    p_vb = (uint16*) (vic_buf);
    vb_end = (uint16*) (vic_buf + 384 * 272);

    do 
    {
      uint16 both;
      for (int x = 0; x < 384; x +=2)   
      {      
	both = *p_vb;      
	*p_fb = ((uint32) rgb565[both >> 8] << 16) + rgb565[both & 0xFF];
	p_fb ++;
	p_vb ++;
      }

      p_fb += 64;
    }
    while (p_vb < vb_end);

  }
  else 
  {
    p_fb = (uint32*) txr_buf[txr_cur];
    p_vb = (uint16*) (vic_buf + 32 + 384 * 16);
    vb_end = (uint16*) (vic_buf + 384 * 256);  

    do 
    {
      uint16 both;
      for (int x = 32; x < 352; x +=2)   
      {      
	both = *p_vb;      
	*p_fb = ((uint32) rgb565[both >> 8] << 16) + rgb565[both & 0xFF];
	p_fb ++;
	p_vb ++;
      }

      p_vb += 32;
      p_fb += 96;
    }
    while (p_vb < vb_end);
  }

}

void C64Display::UploadTexture()
{
  int tex_size;

  if (ThePrefs.ShowBorders)
    tex_size = 512 * 272 * 2;
  else 
    tex_size = 512 * 240 * 2;

  char* ptr = (char*)((unsigned long) txr_buf[txr_cur] + tex_size - 16 * 1024);
  long i = 16 * 1024 / 32;

  do 
  {
    __asm__ __volatile__("ocbp @%0" : : "r"(ptr));
    ptr += 32;
  } 
  while (-- i);

  pvr_txr_load_dma(txr_buf[txr_cur], txr[txr_cur], tex_size, 0);
}

void C64Display::DrawSpeedometer()
{
  char text[10];
  sprintf(text, "%3d%%", speed);

  plx_fcxt_begin(ss_fcxt);
  plx_fcxt_setpos(ss_fcxt, 0.85f * x_size, 0.96f * y_size, 15.0f);
  plx_fcxt_draw(ss_fcxt, text);
  plx_fcxt_end(ss_fcxt);
}

void C64Display::DrawLEDBoxes()
{
  pvr_poly_hdr_t hdr;
  pvr_poly_cxt_t cxt;

  pvr_poly_cxt_col(&cxt, PVR_LIST_OP_POLY);
  cxt.gen.shading = PVR_SHADE_FLAT;
  pvr_poly_compile(&hdr, &cxt);
  pvr_prim(&hdr, sizeof(hdr));

  for (int i = 0; i < 4; i ++)
    if (led_state[3 - i] == LED_ON)
      box_draw_pvr((0.78f - 0.06f * i) * x_size, 0.92f * y_size, 0.05f * x_size, 0.04f * y_size, 15.0f, 2.0f, 
		   1.0f, 0.7f, 0.2f, 0.2f, 1.0f, 0.8f, 0.8f, 0.8f);
    else
      box_draw_pvr((0.78f - 0.06f * i) * x_size, 0.92f * y_size, 0.05f * x_size, 0.04f * y_size, 15.0f, 2.0f,
                   1.0f, 0.3f, 0.1f, 0.1f, 1.0f, 0.8f, 0.8f, 0.8f);
}

void C64Display::DrawKeyboardBoxes()
{
  keyboard_draw_box_pvr(&keybd);
}

void C64Display::DrawKeyboardText()
{
  keyboard_draw_text_pvr(&keybd);
}

void C64Display::DrawLEDText()
{
  plx_fcxt_setsize(fs_fcxt, 12.f);

  for (int i = 0; i < 4; i ++)
  {
    char text[3];

    if (i == 3 && ThePrefs.Emul1541Proc && ThePrefs.DriveType[0] == DRVTYPE_D64)    
      sprintf(text, "%02d", TheApp->TheC64->TheJob1541->current_halftrack);
    else 
      sprintf(text, "%d", 11 - i);

    plx_fcxt_begin(fs_fcxt);
    plx_fcxt_setpos(fs_fcxt, (0.79f - 0.06f * i) * x_size, 0.95f * y_size, 16.0f);
    plx_fcxt_draw(fs_fcxt, text);
    plx_fcxt_end(fs_fcxt);
  }

  plx_fcxt_setsize(fs_fcxt, 18.f);
}

bool C64Display::HandleUserInput()
{
  bool handled = ThePrefs.ShowKeyboard;

  if (controller_check_input(&cntr))
  {
    if (ThePrefs.ShowKeyboard)
    {
      int input = keyboard_operate(&keybd, &cntr.cont.last_st);

      if (input != KEY_NOACTION)
      {
	if (input == MATRIX(1, 7))
	  keybd_shift = !keybd_shift;
	else
	  FakeKeyPress(input, keybd_shift, TheApp->TheC64->TheCIA1->KeyMatrix, TheApp->TheC64->TheCIA1->RevMatrix, 0);
      }
    }

    cont_state_t *cont_state = &(cntr.cont.last_st);

    if ((cont_state->buttons & CONT_X))
    {
      if (ThePrefs.ShowKeyboard)
      {
	if (ThePrefs.KeyboardUp)
	{
	  keyboard_create(&keybd, 0.0f, 240.0f, 640.0f, 240.0f, 10.0f, fs_fnt, fs_fcxt, key_row_list, 9);
	  ThePrefs.ShowKeyboard = false;
	}
	else 
	  keyboard_create(&keybd, 0.0f, 0.0f, 640.0f, 240.0f, 10.0f, fs_fnt, fs_fcxt, key_row_list, 9);
	
	ThePrefs.KeyboardUp = !ThePrefs.KeyboardUp;	       	
      }
      else 
	ThePrefs.ShowKeyboard = true;
    }

    if ((cont_state->buttons & CONT_B))
      FakeKeyPress(ThePrefs.KeyMap[0], keybd_shift, TheApp->TheC64->TheCIA1->KeyMatrix, TheApp->TheC64->TheCIA1->RevMatrix, 0);

    if ((cont_state->buttons & CONT_Y))
      FakeKeyPress(ThePrefs.KeyMap[1], keybd_shift, TheApp->TheC64->TheCIA1->KeyMatrix, TheApp->TheC64->TheCIA1->RevMatrix, 0);

    if ((cont_state->ltrig >= 255))
      FakeKeyPress(ThePrefs.KeyMap[2], keybd_shift, TheApp->TheC64->TheCIA1->KeyMatrix, TheApp->TheC64->TheCIA1->RevMatrix, 0);

    if ((cont_state->rtrig >= 255))
      FakeKeyPress(ThePrefs.KeyMap[3], keybd_shift, TheApp->TheC64->TheCIA1->KeyMatrix, TheApp->TheC64->TheCIA1->RevMatrix, 0);

    if (start_hold != 0)
    {
      if ((cont_state->buttons & CONT_START))
	start_hold --;
      else if (start_hold < 5)
      {
	TheApp->TheC64->Pause();
	
	Prefs newPrefs;
	newPrefs = ThePrefs;
	newPrefs.ShowEditor(false, "");       
	
	TheApp->TheC64->NewPrefs(&newPrefs); 
	ThePrefs = newPrefs;
	
	start_hold = 5;
	TheApp->TheC64->Resume();
      }
    }

    if (start_hold == 1)
    {
      if (ThePrefs.FancyReset)
      {
	anim_type = ANIM_TURN;
	anim_time = 30;
	start_hold = 0;
      }
      else 
	TheApp->TheC64->QuitThyself(); 
    }
  }

  return handled;
}

void C64Display::UpdatePVR()
{
  pvr_wait_ready();
  pvr_scene_begin(); 
  pvr_list_begin(PVR_LIST_OP_POLY);

  Animate();
  CalcCoords();
  SendToPVR(PVR_PACK_COLOR(1.0f, 1.0f, 1.0f, 1.0f));
  if (anim_type == ANIM_NONE)
  {
    if (ThePrefs.ShowLEDs)  
      DrawLEDBoxes();
    if (ThePrefs.ShowKeyboard)
      DrawKeyboardBoxes();
  }
  pvr_list_finish();
  
  if (anim_type == ANIM_NONE)
  {  
    if (ThePrefs.ShowSpdMeter || ThePrefs.ShowLEDs || ThePrefs.ShowKeyboard)
      pvr_list_begin(PVR_LIST_TR_POLY);

    if (ThePrefs.ShowSpdMeter)
      DrawSpeedometer();
    if (ThePrefs.ShowLEDs)
      DrawLEDText();
    if (ThePrefs.ShowKeyboard)
      DrawKeyboardText();

    if (ThePrefs.ShowSpdMeter || ThePrefs.ShowLEDs || ThePrefs.ShowKeyboard)
      pvr_list_finish();
  }
   
  pvr_scene_finish();

}

void C64Display::Update(void)
{
  UploadTexture(); 

  RenderToTexture();

  txr_cur ^= 1;
}

uint8 *C64Display::BitmapBase(void)
{
  return (uint8 *) vic_buf;
}


int C64Display::BitmapXMod(void)
{
  return DISPLAY_X;  
}

void C64Display::PollKeyboard(uint8 *CIA_key_matrix, uint8 *CIA_rev_matrix, uint8 *joystick)
{
  // Clear matrices
  for (int i = 0; i < 8; i ++)
  {
    CIA_key_matrix[i] = 0xFF;
    CIA_rev_matrix[i] = 0xFF;
  }

  // Get raw keyboard status from Maple
  maple_device_t *keybd_dev = maple_enum_type(0, MAPLE_FUNC_KEYBOARD);
  if (keybd_dev == NULL)
    return;   // No Keyboard
  kbd_state_t *keybd_state = (kbd_state_t*) maple_dev_status(keybd_dev);

  if (keybd_state->cond.modifiers & KBD_MOD_LSHIFT)
  {
    CIA_key_matrix[1] &= ~(1 << 7);
    CIA_rev_matrix[7] &= ~(1 << 1);    
  }

  if (keybd_state->cond.modifiers & KBD_MOD_RSHIFT)
  {
    CIA_key_matrix[6] &= 0xef;
    CIA_rev_matrix[4] &= 0xbf;
  }

  if (keybd_state->cond.modifiers & KBD_MOD_LCTRL)
  {
    CIA_key_matrix[7] &= ~(1 << 2);
    CIA_rev_matrix[2] &= ~(1 << 7);    
  }

  if (keybd_state->cond.modifiers & KBD_MOD_RCTRL)
  {
    CIA_key_matrix[7] &= ~(1 << 5);
    CIA_rev_matrix[5] &= ~(1 << 7);    
  }

  for (int i = 0; i < 6; i ++)
  {
    int kc = DreamcastKey2C64(keybd_state->cond.keys[i]); 
        
    if (kc != -1)
    {
      int c64_byte, c64_bit, shifted;
      c64_byte = kc >> 3;
      c64_bit = kc & 7;
      shifted = kc & 128;
      c64_byte &= 7;
      if (shifted) 
      {
	CIA_key_matrix[6] &= 0xef;
	CIA_rev_matrix[4] &= 0xbf;
      }
      CIA_key_matrix[c64_byte] &= ~(1 << c64_bit);
      CIA_rev_matrix[c64_bit] &= ~(1 << c64_byte);
    }
  }
}

void C64Display::FakeKeyPress(int kc, bool shift, uint8 *CIA_key_matrix, uint8 *CIA_rev_matrix, uint8 *joystick)
{
  // Clear matrices
  for (int i = 0; i < 8; i ++)
  {
    CIA_key_matrix[i] = 0xFF;
    CIA_rev_matrix[i] = 0xFF;
  }

  if (shift)
  {
    CIA_key_matrix[6] &= 0xef;
    CIA_rev_matrix[4] &= 0xbf;
  }

  if (kc != -1)
  {
    int c64_byte, c64_bit, shifted;
    c64_byte = kc >> 3;
    c64_bit = kc & 7;
    shifted = kc & 128;
    c64_byte &= 7;
    if (shifted) 
    {
      CIA_key_matrix[6] &= 0xef;
      CIA_rev_matrix[4] &= 0xbf;
    }
    CIA_key_matrix[c64_byte] &= ~(1 << c64_bit);
    CIA_rev_matrix[c64_bit] &= ~(1 << c64_byte);
  }
}

/*
 *  Check if NumLock is down (for switching the joystick keyboard emulation)
 */

bool C64Display::NumLock(void)
{
       return false; //numlock;
}

void C64Display::InitColors(uint8 *colors)
{
  for (int i = 0; i < 256; i ++) 
    colors[i] = i & 0x0F;

  for (int i = 0; i < 16; i ++)
    rgb565[i] = ((palette_red[i] >> 3) << 11) | ((palette_green[i] >> 3) << 6) | ((palette_blue[i] >> 3) << 0);
}

/*
 *  Show a requester (error message)
 */

long int ShowRequester(char *a, char *b, char *c)
{
  requester_t req;
  
  menu_item_t choice_items[] = {
    { b, 1, 0, 0 },
    { c, 0, 0, 0 },
  };   

  int choice_cnt = 1;
  if (c)
    choice_cnt = 2;

  requester_create(&req, 80.0f, 160.0f, 480.f, 80.0f, 120.0f, 10.0f + 30.0f * choice_cnt, 11.0f, a, choice_items, choice_cnt, 
		   TheApp->TheC64->TheDisplay->ss_fnt, TheApp->TheC64->TheDisplay->ss_fcxt);

  req.msg_menu.col.in.r = 0.5f;
  req.msg_menu.col.in.g = 0.1f;
  req.msg_menu.col.in.b = 0.1f;
  req.msg_menu.items.sel = -1;

  int input = MENU_NOACTION;

  controller_t cntr; 
  controller_create(&cntr, 0, 20);

  do 
  {
    pvr_wait_ready();
    //while (pvr_check_ready() == -1) { };
    pvr_scene_begin();

    pvr_list_begin(PVR_LIST_OP_POLY);
    TheApp->TheC64->TheDisplay->SendToPVR(PVR_PACK_COLOR(1.0f, 0.5f, 0.5f, 0.5f));
    menu_draw_box_chain_pvr(&(req.msg_menu));
    menu_draw_box_chain_pvr(&(req.choice_menu));
    pvr_list_finish();

    pvr_list_begin(PVR_LIST_TR_POLY);
    menu_draw_text_chain_pvr(&(req.msg_menu));
    menu_draw_text_chain_pvr(&(req.choice_menu));
    pvr_list_finish();

    pvr_scene_finish();
    
    if (controller_check_input(&cntr))
      input = menu_operate(&(req.choice_menu), &cntr.cont.last_st);
  } while (input == MENU_NOACTION || input == MENU_CLOSE);
  
  return input;
}

void ShowMessage(char *msg)
{
  pvr_poly_hdr_t hdr;
  pvr_poly_cxt_t cxt;

  plx_font_t *ss_fnt;
  plx_fcxt_t *ss_fcxt;
  
  ss_fnt = TheApp->TheC64->TheDisplay->ss_fnt;
  ss_fcxt = TheApp->TheC64->TheDisplay->ss_fcxt;
  
  pvr_wait_ready();
  //  while (pvr_check_ready() == -1) { };
  pvr_scene_begin();

  pvr_list_begin(PVR_LIST_OP_POLY);

  TheApp->TheC64->TheDisplay->SendToPVR(PVR_PACK_COLOR(1.0f, 0.5f, 0.5f, 0.5f));

  pvr_poly_cxt_col(&cxt, PVR_LIST_OP_POLY);
  cxt.gen.shading = PVR_SHADE_FLAT;
  pvr_poly_compile(&hdr, &cxt);
  pvr_prim(&hdr, sizeof(hdr));
  
  box_draw_pvr(80.0f, 160.0f, 480.f, 80.0f, 11.0f, 2.0f, 
	       1.0f, 0.5f, 0.1f, 0.1f, 1.0f, 0.8f, 0.8f, 0.8f);

  pvr_list_finish();
  
  pvr_list_begin(PVR_LIST_TR_POLY);

  plx_fcxt_begin(ss_fcxt);
  plx_fcxt_setpos(ss_fcxt, 108.0f, 210.0f, 12.0f);
  plx_fcxt_draw(ss_fcxt, msg);
  plx_fcxt_end(ss_fcxt);

  pvr_list_finish();
  pvr_scene_finish();
}
